from collections.abc import Callable
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pydantic_core._pydantic_core import PydanticCustomError


# NOTE: We use the factory approach because PydanticCustomError is a final class.
#   That is also why this module is internal.


def CustomError(fn: Callable, invalid_tag: str, **kwargs) -> "PydanticCustomError":
    # perf: keep module loading super fast by localizing this import.
    from pydantic_core._pydantic_core import PydanticCustomError

    return PydanticCustomError(fn.__name__, f"Invalid {invalid_tag}", kwargs)


def HexValueError(value: Any) -> "PydanticCustomError":
    return CustomError(HexValueError, "hex value", value=value)


def SizeError(size: Any, value: Any) -> "PydanticCustomError":
    return CustomError(SizeError, "size of value", size=size, value=value)


def Bip122UriFormatError(value: str) -> "PydanticCustomError":
    return CustomError(
        Bip122UriFormatError,
        "BIP-122 URI format",
        uri=value,
        format="blockchain://<genesis_hash>/block/<block_hash>.",
    )



from pydantic import BaseModel
from pydantic_core.core_schema import str_schema, with_info_before_validator_function

from eth_pydantic_types import HexStr32


class MyAddress(HexStr32):
    @classmethod
    def __get_pydantic_core_schema__(cls, value, handler=None):
        str_size = cls.size * 2
        return with_info_before_validator_function(
            cls.__eth_pydantic_validate__,
            str_schema(max_length=str_size, min_length=str_size),
        )

    @classmethod
    def __eth_pydantic_validate__(cls, value, info=None, **kwargs):
        return super().__eth_pydantic_validate__(value, info=info, prefixed=False, **kwargs)


class MyModel(BaseModel):
    address: MyAddress


class TestCustomTypes:
    def test_unprefixed_from_str(self):
        model = MyModel(address="0x" + "ab" * 32)
        assert not model.address.startswith("0x")
        assert len(model.address) == 64

    def test_unprefixed_from_bytes(self):
        model = MyModel(address=b"\xab" * 32)
        assert not model.address.startswith("0x")
        assert len(model.address) == 64

    def test_unprefixed_from_int(self):
        model = MyModel(address=123)
        assert not model.address.startswith("0x")
        assert len(model.address) == 64



"""
These models are used to match the lowercase type names used by the abi.
"""

from typing import Annotated, ClassVar

from pydantic import Field
from typing_extensions import TypeAliasType

from .address import Address
from .hex import BoundHexBytes, HexBytes

bool = TypeAliasType("bool", bool)
bytes = TypeAliasType("bytes", HexBytes)
string = TypeAliasType("string", str)
address = TypeAliasType("address", Address)


class bytes1(BoundHexBytes):
    size: ClassVar[int] = 1


class bytes2(BoundHexBytes):
    size: ClassVar[int] = 2


class bytes3(BoundHexBytes):
    size: ClassVar[int] = 3


class bytes4(BoundHexBytes):
    size: ClassVar[int] = 4


class bytes5(BoundHexBytes):
    size: ClassVar[int] = 5


class bytes6(BoundHexBytes):
    size: ClassVar[int] = 6


class bytes7(BoundHexBytes):
    size: ClassVar[int] = 7


class bytes8(BoundHexBytes):
    size: ClassVar[int] = 8


class bytes9(BoundHexBytes):
    size: ClassVar[int] = 9


class bytes10(BoundHexBytes):
    size: ClassVar[int] = 10


class bytes11(BoundHexBytes):
    size: ClassVar[int] = 11


class bytes12(BoundHexBytes):
    size: ClassVar[int] = 12


class bytes13(BoundHexBytes):
    size: ClassVar[int] = 13


class bytes14(BoundHexBytes):
    size: ClassVar[int] = 14


class bytes15(BoundHexBytes):
    size: ClassVar[int] = 15


class bytes16(BoundHexBytes):
    size: ClassVar[int] = 16


class bytes17(BoundHexBytes):
    size: ClassVar[int] = 17


class bytes18(BoundHexBytes):
    size: ClassVar[int] = 18


class bytes19(BoundHexBytes):
    size: ClassVar[int] = 19


class bytes20(BoundHexBytes):
    size: ClassVar[int] = 20


class bytes21(BoundHexBytes):
    size: ClassVar[int] = 21


class bytes22(BoundHexBytes):
    size: ClassVar[int] = 22


class bytes23(BoundHexBytes):
    size: ClassVar[int] = 23


class bytes24(BoundHexBytes):
    size: ClassVar[int] = 24


class bytes25(BoundHexBytes):
    size: ClassVar[int] = 25


class bytes26(BoundHexBytes):
    size: ClassVar[int] = 26


class bytes27(BoundHexBytes):
    size: ClassVar[int] = 27


class bytes28(BoundHexBytes):
    size: ClassVar[int] = 28


class bytes29(BoundHexBytes):
    size: ClassVar[int] = 29


class bytes30(BoundHexBytes):
    size: ClassVar[int] = 30


class bytes31(BoundHexBytes):
    size: ClassVar[int] = 31


class bytes32(BoundHexBytes):
    size: ClassVar[int] = 32


int8 = TypeAliasType("int8", Annotated[int, Field(lt=2**7, ge=-(2**7))])
int16 = TypeAliasType("int16", Annotated[int, Field(lt=2**15, ge=-(2**15))])
int24 = TypeAliasType("int24", Annotated[int, Field(lt=2**23, ge=-(2**23))])
int32 = TypeAliasType("int32", Annotated[int, Field(lt=2**31, ge=-(2**31))])
int40 = TypeAliasType("int40", Annotated[int, Field(lt=2**39, ge=-(2**39))])
int48 = TypeAliasType("int48", Annotated[int, Field(lt=2**47, ge=-(2**47))])
int56 = TypeAliasType("int56", Annotated[int, Field(lt=2**55, ge=-(2**55))])
int64 = TypeAliasType("int64", Annotated[int, Field(lt=2**63, ge=-(2**63))])
int72 = TypeAliasType("int72", Annotated[int, Field(lt=2**71, ge=-(2**71))])
int80 = TypeAliasType("int80", Annotated[int, Field(lt=2**79, ge=-(2**79))])
int88 = TypeAliasType("int88", Annotated[int, Field(lt=2**87, ge=-(2**87))])
int96 = TypeAliasType("int96", Annotated[int, Field(lt=2**95, ge=-(2**95))])
int104 = TypeAliasType("int104", Annotated[int, Field(lt=2**103, ge=-(2**103))])
int112 = TypeAliasType("int112", Annotated[int, Field(lt=2**111, ge=-(2**111))])
int120 = TypeAliasType("int120", Annotated[int, Field(lt=2**119, ge=-(2**119))])
int128 = TypeAliasType("int128", Annotated[int, Field(lt=2**127, ge=-(2**127))])
int136 = TypeAliasType("int136", Annotated[int, Field(lt=2**135, ge=-(2**135))])
int144 = TypeAliasType("int144", Annotated[int, Field(lt=2**143, ge=-(2**143))])
int152 = TypeAliasType("int152", Annotated[int, Field(lt=2**151, ge=-(2**151))])
int160 = TypeAliasType("int160", Annotated[int, Field(lt=2**159, ge=-(2**159))])
int168 = TypeAliasType("int168", Annotated[int, Field(lt=2**167, ge=-(2**167))])
int176 = TypeAliasType("int176", Annotated[int, Field(lt=2**175, ge=-(2**175))])
int184 = TypeAliasType("int184", Annotated[int, Field(lt=2**183, ge=-(2**183))])
int192 = TypeAliasType("int192", Annotated[int, Field(lt=2**191, ge=-(2**191))])
int200 = TypeAliasType("int200", Annotated[int, Field(lt=2**199, ge=-(2**199))])
int208 = TypeAliasType("int208", Annotated[int, Field(lt=2**207, ge=-(2**207))])
int216 = TypeAliasType("int216", Annotated[int, Field(lt=2**215, ge=-(2**215))])
int224 = TypeAliasType("int224", Annotated[int, Field(lt=2**223, ge=-(2**223))])
int232 = TypeAliasType("int232", Annotated[int, Field(lt=2**231, ge=-(2**231))])
int240 = TypeAliasType("int240", Annotated[int, Field(lt=2**239, ge=-(2**239))])
int248 = TypeAliasType("int248", Annotated[int, Field(lt=2**247, ge=-(2**247))])
int256 = TypeAliasType("int256", Annotated[int, Field(lt=2**255, ge=-(2**255))])
uint8 = TypeAliasType("uint8", Annotated[int, Field(lt=2**8, ge=0)])
uint16 = TypeAliasType("uint16", Annotated[int, Field(lt=2**16, ge=0)])
uint24 = TypeAliasType("uint24", Annotated[int, Field(lt=2**24, ge=0)])
uint32 = TypeAliasType("uint32", Annotated[int, Field(lt=2**32, ge=0)])
uint40 = TypeAliasType("uint40", Annotated[int, Field(lt=2**40, ge=0)])
uint48 = TypeAliasType("uint48", Annotated[int, Field(lt=2**48, ge=0)])
uint56 = TypeAliasType("uint56", Annotated[int, Field(lt=2**56, ge=0)])
uint64 = TypeAliasType("uint64", Annotated[int, Field(lt=2**64, ge=0)])
uint72 = TypeAliasType("uint72", Annotated[int, Field(lt=2**72, ge=0)])
uint80 = TypeAliasType("uint80", Annotated[int, Field(lt=2**80, ge=0)])
uint88 = TypeAliasType("uint88", Annotated[int, Field(lt=2**88, ge=0)])
uint96 = TypeAliasType("uint96", Annotated[int, Field(lt=2**96, ge=0)])
uint104 = TypeAliasType("uint104", Annotated[int, Field(lt=2**104, ge=0)])
uint112 = TypeAliasType("uint112", Annotated[int, Field(lt=2**112, ge=0)])
uint120 = TypeAliasType("uint120", Annotated[int, Field(lt=2**120, ge=0)])
uint128 = TypeAliasType("uint128", Annotated[int, Field(lt=2**128, ge=0)])
uint136 = TypeAliasType("uint136", Annotated[int, Field(lt=2**136, ge=0)])
uint144 = TypeAliasType("uint144", Annotated[int, Field(lt=2**144, ge=0)])
uint152 = TypeAliasType("uint152", Annotated[int, Field(lt=2**152, ge=0)])
uint160 = TypeAliasType("uint160", Annotated[int, Field(lt=2**160, ge=0)])
uint168 = TypeAliasType("uint168", Annotated[int, Field(lt=2**168, ge=0)])
uint176 = TypeAliasType("uint176", Annotated[int, Field(lt=2**176, ge=0)])
uint184 = TypeAliasType("uint184", Annotated[int, Field(lt=2**184, ge=0)])
uint192 = TypeAliasType("uint192", Annotated[int, Field(lt=2**192, ge=0)])
uint200 = TypeAliasType("uint200", Annotated[int, Field(lt=2**200, ge=0)])
uint208 = TypeAliasType("uint208", Annotated[int, Field(lt=2**208, ge=0)])
uint216 = TypeAliasType("uint216", Annotated[int, Field(lt=2**216, ge=0)])
uint224 = TypeAliasType("uint224", Annotated[int, Field(lt=2**224, ge=0)])
uint232 = TypeAliasType("uint232", Annotated[int, Field(lt=2**232, ge=0)])
uint240 = TypeAliasType("uint240", Annotated[int, Field(lt=2**240, ge=0)])
uint248 = TypeAliasType("uint248", Annotated[int, Field(lt=2**248, ge=0)])
uint256 = TypeAliasType("uint256", Annotated[int, Field(lt=2**256, ge=0)])



from typing import TYPE_CHECKING, Any, ClassVar

from hexbytes.main import HexBytes as BaseHexBytes
from pydantic_core.core_schema import (
    ValidationInfo,
    no_info_before_validator_function,
    str_schema,
    with_info_before_validator_function,
)

from eth_pydantic_types._error import HexValueError
from eth_pydantic_types.hex.base import BaseHex
from eth_pydantic_types.utils import (
    PadDirection,
    get_hash_examples,
    get_hash_pattern,
    validate_hex_str,
    validate_str_size,
)

if TYPE_CHECKING:
    from pydantic_core import CoreSchema
    from typing_extensions import TypeAlias


class BaseHexStr(str, BaseHex):
    @classmethod
    def __get_pydantic_core_schema__(cls, value, handler=None):
        return no_info_before_validator_function(cls.__eth_pydantic_validate__, str_schema())

    @classmethod
    def __eth_pydantic_validate__(cls, value, **kwargs):
        return value  # Override.

    @classmethod
    def from_bytes(cls, data: bytes) -> "BaseHexStr":
        hex_value = data.hex()
        hex_str = hex_value if hex_value.startswith("0x") else f"0x{hex_value}"
        return cls(hex_str)

    @classmethod
    def validate_hex(cls, data: bytes | str | int, prefixed: bool = True):
        if isinstance(data, bytes):
            result = cls.from_bytes(data)
        elif isinstance(data, str):
            result = cls(validate_hex_str(data))
        elif isinstance(data, int):
            hex_value = BaseHexBytes(data).hex()
            result = cls(hex_value if hex_value.startswith("0x") else f"0x{hex_value}")
        else:
            raise HexValueError(data)

        if not prefixed and result.startswith("0x"):
            return result[2:]

        return result

    def __int__(self) -> int:
        return int(self, 16)

    def __bytes__(self) -> bytes:
        return bytes.fromhex(self[2:])


class HexStr(BaseHexStr):
    """A hex string value."""

    @classmethod
    def __get_pydantic_core_schema__(cls, value, handler=None) -> "CoreSchema":
        return with_info_before_validator_function(
            cls.__eth_pydantic_validate__,
            str_schema(),
        )

    @classmethod
    def __eth_pydantic_validate__(
        cls, value: Any, info: ValidationInfo | None = None, **kwargs
    ) -> str:
        hex_str = cls.validate_hex(value)
        hex_value = hex_str[2:] if hex_str.startswith("0x") else hex_str
        return cls(f"0x{hex_value}")

    @classmethod
    def from_bytes(cls, data: bytes) -> "HexStr":
        value_str = super().from_bytes(data)
        value = value_str if value_str.startswith("0x") else f"0x{value_str}"
        return HexStr(value)


class BoundHexStr(BaseHexStr):
    """A hex string value, that is required to be a specific size."""

    size: ClassVar[int] = 32

    @classmethod
    def __get_pydantic_core_schema__(cls, value, handler=None) -> "CoreSchema":
        str_size = cls.size * 2 + 2
        return with_info_before_validator_function(
            cls.__eth_pydantic_validate__,
            str_schema(max_length=str_size, min_length=str_size),
        )

    @classmethod
    def __eth_pydantic_validate__(
        cls, value: Any, info: ValidationInfo | None = None, **kwargs
    ) -> str:
        if not (pad := kwargs.pop("pad", None)):
            # Integers are always padded to the left, but bytes-types are padded to the right
            # to be ABI-encode compliant.
            pad = PadDirection.LEFT if isinstance(value, int) else PadDirection.RIGHT

        prefixed = kwargs.pop("prefixed", True)
        hex_str = cls.validate_hex(value, prefixed=False)
        sized_value = cls.validate_size(hex_str, pad_direction=pad)
        return cls(f"0x{sized_value}") if prefixed else cls(sized_value)

    @classmethod
    def validate_size(cls, value: str, pad_direction: PadDirection = PadDirection.LEFT) -> str:
        cls.update_schema()
        return validate_str_size(value, cls.size * 2, pad_direction=pad_direction)

    @classmethod
    def update_schema(cls):
        str_size = cls.size * 2
        cls.schema_pattern = get_hash_pattern(str_size)
        cls.schema_examples = get_hash_examples(str_size)


class HexStr20(BoundHexStr):
    size: ClassVar[int] = 20


HexStr32: "TypeAlias" = BoundHexStr



import pytest
from pydantic import BaseModel, ValidationError

from eth_pydantic_types.bip122 import Bip122Uri, Bip122UriType

GENESIS_HASH = "d4e56740f876aef8c010b86a40d5f56745a118d0906a34e69aec8c0db1cb8fa3"
BLOCK_HASH = "752820c0ad7abc1200f9ad42c4adc6fbb4bd44b5bed4667990e64565102c1ba6"
EXAMPLE = f"blockchain://{GENESIS_HASH}/block/{BLOCK_HASH}"


class Model(BaseModel):
    uri: Bip122Uri


def test_parsed():
    # NOTE: This won't work in the model itself because
    #   it turns to `str` after validation, much like URL pydantic classes.
    uri = Bip122Uri(EXAMPLE)
    assert uri == EXAMPLE
    assert uri.uri_type == Bip122UriType.BLOCK
    assert uri.chain == f"0x{GENESIS_HASH}"
    assert uri.hash == f"0x{BLOCK_HASH}"


@pytest.mark.parametrize(
    "uri",
    (
        "foo",
        f"blockchain://foo/block/{BLOCK_HASH}",
        f"blockchain://{GENESIS_HASH}/block/foo",
        f"blockchain://{GENESIS_HASH}/tx/foo",
    ),
)
def test_invalid_bip122(uri):
    with pytest.raises(ValidationError):
        Model(uri=uri)


def test_schema():
    actual = Model.model_json_schema()
    prop = actual["properties"]["uri"]
    assert prop["type"] == "string"
    assert prop["title"] == "Uri"
    assert prop["examples"] == [EXAMPLE]
    assert prop["pattern"] == "^blockchain://[0-9a-f]{64}/block/[0-9a-f]{64}$"



import pytest
from pydantic import BaseModel, ValidationError

from eth_pydantic_types.address import Address, AddressType
from eth_pydantic_types.hex import HexBytes

# NOTE: This address purposely is the wrong length (missing left zero),
#   not checksummed, and not 0x prefixed.
ADDRESS = "837207e343277cbd6c114a45ec0e9ec56a1ad84"
CHECKSUM_ADDRESS = "0x0837207e343277CBd6c114a45EC0e9Ec56a1AD84"

LEADING_ZEROES_ADDRESS = "04444c5dc75cb358380d2e3de08a90"  # Uniswap v4 pool manager
CHECKSUM_LEADING_ZEROES_ADDRESS = "0x000000000004444c5dc75cB358380D2e3dE08A90"


class Model(BaseModel):
    address: Address
    address_type: AddressType


@pytest.fixture
def checksum_address():
    return CHECKSUM_ADDRESS


@pytest.mark.parametrize(
    "address",
    (
        CHECKSUM_ADDRESS,
        ADDRESS,
        f"0x{ADDRESS}",
        f"0x0{ADDRESS}",
        int(ADDRESS, 16),
        HexBytes(ADDRESS),
    ),
)
def test_address(address, checksum_address):
    actual = Model(address=address, address_type=address)
    assert actual.address == checksum_address
    assert actual.address_type == checksum_address


@pytest.mark.parametrize("address", ("foo", -35, "0x" + ("F" * 100)))
def test_invalid_address(address):
    with pytest.raises(ValidationError):
        Model(address=address, address_type=address)


def test_schema():
    actual = Model.model_json_schema()
    prop = actual["properties"]["address"]
    assert prop["maxLength"] == 42
    assert prop["minLength"] == 42
    assert prop["type"] == "string"
    assert prop["pattern"] == "^0x[a-fA-F0-9]{40}$"
    assert prop["examples"] == [
        "0x0000000000000000000000000000000000000000",
        "0x02c84e944F97F4A4f60221e6fb5d5DbAE49c7aaB",
        "0xa5a13f62ce1113838e0d9b4559b8caf5f76463c0",
        "0x1e59ce931B4CFea3fe4B875411e280e173cB7A9C",
    ]


def test_model_dump():
    model = Model(address=ADDRESS, address_type=ADDRESS)
    actual = model.model_dump()
    expected = {"address": CHECKSUM_ADDRESS, "address_type": CHECKSUM_ADDRESS}
    assert actual == expected


def test_leading_zeroes():
    model = Model(address=LEADING_ZEROES_ADDRESS, address_type=LEADING_ZEROES_ADDRESS)
    actual = model.model_dump()
    expected = {
        "address": CHECKSUM_LEADING_ZEROES_ADDRESS,
        "address_type": CHECKSUM_LEADING_ZEROES_ADDRESS,
    }
    assert actual == expected



from typing import ClassVar

schema_pattern = "^0x([0-9a-f][0-9a-f])*$"
schema_examples = (
    "0x",  # empty bytes
    "0xd4",
    "0xd4e5",
    "0xd4e56740",
    "0xd4e56740f876aef8",
    "0xd4e56740f876aef8c010b86a40d5f567",
    "0xd4e56740f876aef8c010b86a40d5f56745a118d0906a34e69aec8c0db1cb8fa3",
)


class BaseHex:
    size: ClassVar[int] = 0
    schema_pattern: ClassVar[str] = schema_pattern
    schema_examples: ClassVar[tuple[str, ...]] = schema_examples

    @classmethod
    def __get_pydantic_json_schema__(cls, core_schema, handler):
        json_schema = handler(core_schema)
        json_schema.update(
            format="binary",
            pattern=cls.schema_pattern,
            examples=list(cls.schema_examples),
        )
        return json_schema



from pydantic_core.core_schema import plain_serializer_function_ser_schema

from eth_pydantic_types.utils import PadDirection, validate_str_size


def serialize_hex(
    value: int | bytes,
    size: int | None = None,
    pad: PadDirection | None = None,
    force_even_length: bool | None = None,
) -> str:
    hex_value = value.hex() if isinstance(value, bytes) else hex(value)
    hex_value = hex_value[2:] if hex_value.startswith("0x") else hex_value

    if pad is not None and force_even_length is not None:
        # If padding at all and force_even_length not specified, assume True.
        force_even_length = True

    # Ensure even number of left-padded zeroes.
    if force_even_length and len(hex_value) % 2 != 0:
        hex_value = f"0{hex_value}"

    if size is not None and pad is not None:
        hex_value = validate_str_size(hex_value, size * 2, pad_direction=pad)

    return f"0x{hex_value}"


def create_hex_serializer(
    size: int | None = None,
    pad: PadDirection | None = None,
    force_even_length: bool | None = None,
):
    return plain_serializer_function_ser_schema(
        function=lambda value: serialize_hex(
            value, size=size, pad=pad, force_even_length=force_even_length
        )
    )


hex_serializer = create_hex_serializer()



def __getattr__(name: str):
    if "str" in name.lower():
        import eth_pydantic_types.hex.str

        return getattr(eth_pydantic_types.hex.str, name)

    elif "int" in name.lower():
        import eth_pydantic_types.hex.int

        return getattr(eth_pydantic_types.hex.int, name)

    # bytes.
    import eth_pydantic_types.hex.bytes

    return getattr(eth_pydantic_types.hex.bytes, name)
